function checkDirKey(key) --returns if a directional key (up,down,left,right) or it's corresponding WASD key is pressed
  local keyAlts = {['left']='a',['right']='d',['up']='w',['down']='s'}
  return love.keyboard.isDown(key,keyAlts[key])
end
function checkBBCollision(x1,y1,w1,h1,x2,y2,w2,h2,fromcenter)--checks collision with two boxes
  if fromcenter == nil then fromcenter = false end
  if not fromcenter then
    return x1 < x2+w2 and x2 < x1+w1 and y1 < y2+h2 and y2 < y1+h1
  else
    return x1 < x2+w2/2 and x2 < x1+w1/2 and y1 < y2+h2/2 and y2 < y1+h1/2
  end
end
function tableLength(T)--returns the length of a table, duuuuh
  local count = 0
  for i,_ in ipairs(T) do count = count + 1 end
  return count
end
function pointCollision(x1,y1,x2,y2,w2,h2,fromcenter)--checks if a point is inside a box
  if fromcenter == nil then fromcenter = false end
  if not fromcenter then
    return x1 > x2 and x1 < x2+w2 and y1 > y2 and y1 < y2+h2
  else
    return x1 > x2-w2 and x1 < x2+w2/2 and y1 > y2-h2 and y1 < y2+h2/2
  end
end
function randomChance(percent,invert)--random chances
  local randNum = math.random(0,100)
  if invert == nil then invert = false end
  if not invert then return randNum < percent else return randNum > percent end
end
function newRand(min,max,dec)--returns a random number, but with more decimals (the default math.random function only returns whole numbers)
  if dec == nil then dec = 2 end
  local toReturn = tostring(math.random(min,max)) .. "."
  for i=1,dec do toReturn = toReturn .. tostring(math.random(0,8)) end
  return tonumber(toReturn)
end
function randomChanceDec(percent,invert,dec)--random chances with decimals
  if dec == nil then dec = 3 end
  local randNum = newRand(0,100,dec)
  if invert == nil then invert = false end
  if not invert then return randNum < percent else return randNum > percent end
end
function degToRad(deg)--converts degrees to radians
  return math.pi/180 * deg
end
function randElement(T)--returns a random element in a table
  return T[math.random(1,tableLength(T))]
end

player = {x=love.graphics.getWidth()/2,--the player metatable
  y=love.graphics.getHeight()/2,
  shopVals = {
    walkSpeed={value=4,name="Moving Speed"},
    maxHealth={value=30,name="Maximum Health"},
    attack={value=1,name="Bullet Power"}
  },
  isMoving=false,
  angle=0,
  moveAngle=0,
  health=30,
  bullets={},canShoot=false,shootTimer=0,shootTimerMax=1.5,
  img=love.graphics.newImage("images/playership.png")
}

shieldImg = love.graphics.newImage("images/shield.png")

bulletImg = love.graphics.newImage("images/bullet.png")--bullet image used by the player AND the enemies(actual image is white, making it colorable)
bullets = {}--stores the bullets that have been fired

enemies = {}--array to store enemies
addEnemyTimer = 0
addEnemyTimerMax = 15--how often enemies should be added(smaller number means more enemies)
enemyImg = love.graphics.newImage("images/enemyship.png")--image used to draw enemy

healthBoxes = {}--array to store health boxes
hlthBoxImg = love.graphics.newImage("images/healthbox.png")--image used to draw health box

score = 0--the score wich shows how many enemies have been killed

gameOver = false--is game over
gamePaused = false--is game paused

gameStarted = false--should i really explain this?

cursorImg = love.graphics.newImage("images/cursor.png")--image used to draw the mouse cursor

stars = {}--contains all of the "stars"

powerupAv = false
isPowerup = false
powerups = {"slow","shield"}
powerupsStrings = {["slow"]="Slow Motion",["shield"]="Shield"}
curPowerup = randElement(powerups)
powerupTimer = 0
powerupTimerMax = 20

timeMult = 50 --only used for convenience, if you want to change the game speed then change the two variables defined below
gameSpeed = 1 --speed of everything including player and enemies and bullets and pretty much everything
worldSpeed = 1 --speed of everything except for the player and his bullets

hurtSounds = {}
for i=1,5 do table.insert(hurtSounds,love.audio.newSource("sounds/hurt" .. i .. ".wav")) end
shootSounds = {}
for i=1,5 do table.insert(shootSounds,love.audio.newSource("sounds/shoot" .. i .. ".wav")) end
healSound = love.audio.newSource("sounds/heal.mp3")

function love.load()
  math.randomseed(os.time())--random seed for the random generators
  curPowerup = randElement(powerups)--random next powerup
  love.mouse.setVisible(false)--hides the mouse
  for i=1,25 do
    table.insert(stars,{x=math.random(0,love.graphics.getWidth()),y=math.random(0,love.graphics.getHeight())})--make "stars"
  end
end

function love.update(dt)
  if not gameOver then
    if not gamePaused and gameStarted then
      --main game logic here
      if checkDirKey('left') or checkDirKey('right') or checkDirKey('up') or checkDirKey('down') then
        player.isMoving = true
        if checkDirKey('down') then--code that checks movement keys is pretty long to normalize diagonal movement
          if checkDirKey('down') and checkDirKey('left') then
            player.moveAngle = degToRad(225+270)
          elseif checkDirKey('down') and checkDirKey('right') then
            player.moveAngle = degToRad(135+270)
          else
            player.moveAngle = degToRad(180+270)
          end
        elseif checkDirKey('up') then
          if checkDirKey('up') and checkDirKey('left') then
            player.moveAngle = degToRad(315+270)
          elseif checkDirKey('up') and checkDirKey('right') then
            player.moveAngle = degToRad(45+270)
          else
            player.moveAngle = degToRad(360+270)
          end
        elseif checkDirKey('left') then
          player.moveAngle = degToRad(270+270)
        elseif checkDirKey('right') then
          player.moveAngle = degToRad(90+270)
        end
      else
        player.isMoving = false
      end
      if player.isMoving then
        player.x = player.x + math.cos(player.moveAngle)*player.shopVals.walkSpeed.value * dt * timeMult * gameSpeed
        player.y = player.y + math.sin(player.moveAngle)*player.shopVals.walkSpeed.value * dt * timeMult * gameSpeed
      end
      if player.x > love.graphics.getWidth() then player.x = love.graphics.getWidth() elseif player.x < 0 then player.x = 0 end--protect the player from moving out of the screen
      if player.y > love.graphics.getHeight() then player.y = love.graphics.getHeight() elseif player.y < 0 then player.y = 0 end--protect the player from moving out of the screen
      player.shootTimer = player.shootTimer + 0.1 * dt * timeMult * gameSpeed
      if player.shootTimer > player.shootTimerMax then
        player.shootTimer = 0
        player.canShoot = true
      else
        player.canShoot = false
      end
      if love.mouse.isDown(1) and player.canShoot then
        --player shoots here
        table.insert(bullets,{x=player.x,y=player.y,angle=player.angle,fromPlayer=true,speed=6,c={r=0,g=255,b=255}})
        love.audio.play(randElement(shootSounds))
      end
      if addEnemyTimer < addEnemyTimerMax then addEnemyTimer = addEnemyTimer + 0.1 * dt * timeMult * gameSpeed * worldSpeed elseif randomChance(1) then
        --adds enemy here
        local newEnemy = {x=math.random(0,love.graphics.getWidth()),y=math.random(0,love.graphics.getHeight()),shootTimer=0,shootTimerMax=math.random(5,10),health=0,maxHealth=math.floor(math.random(15,25)),angle=0,speed=1}
        newEnemy.health = newEnemy.maxHealth
        newEnemy.angle = math.atan2(player.y-newEnemy.y,player.x-newEnemy.x)
        table.insert(enemies,newEnemy)
        addEnemyTimer = 0
      end
      for i,enemy in pairs(enemies) do
        --controls all enemies here
        enemy.angle = math.atan2(player.y-enemy.y,player.x-enemy.x)
        enemy.x = enemy.x + math.cos(enemy.angle)*enemy.speed * dt * timeMult * gameSpeed * worldSpeed
        enemy.y = enemy.y + math.sin(enemy.angle)*enemy.speed * dt * timeMult * gameSpeed * worldSpeed
        if enemy.shootTimer < enemy.shootTimerMax then enemy.shootTimer = enemy.shootTimer + 0.1 * dt * timeMult * gameSpeed * worldSpeed else
          --enemy shoots here
          table.insert(bullets,{x=enemy.x,y=enemy.y,angle=enemy.angle,speed=3,fromPlayer=false,c={r=255,g=0,b=0}})
          enemy.shootTimer = 0
          love.audio.play(love.audio.newSource("sounds/shoot" .. math.random(1,5) .. ".wav"))
        end
        for j,bullet in ipairs(bullets) do
          if pointCollision(bullet.x,bullet.y,enemy.x,enemy.y,enemyImg:getWidth(),enemyImg:getHeight(),true) and bullet.fromPlayer == true then
            --if one of the player's bullets touches an enemy
            enemy.health = enemy.health - player.shopVals.attack.value
            love.audio.play(randElement(hurtSounds))
            table.remove(bullets,j)
          end
        end
        if enemy.health <= 0 then
          --enemy dies here
          table.remove(enemies,i)
          score = score + 1
        end
      end
      for j,bullet in pairs(bullets) do
        --controls all bullets here
        if bullet.fromPlayer then
          bullet.x = bullet.x + math.cos(bullet.angle)*bullet.speed * dt * timeMult * gameSpeed
          bullet.y = bullet.y + math.sin(bullet.angle)*bullet.speed * dt * timeMult * gameSpeed
        else
          bullet.x = bullet.x + math.cos(bullet.angle)*bullet.speed * dt * timeMult * gameSpeed * worldSpeed
          bullet.y = bullet.y + math.sin(bullet.angle)*bullet.speed * dt * timeMult * gameSpeed * worldSpeed
        end
        if pointCollision(bullet.x,bullet.y,player.x,player.y,player.img:getWidth(),player.img:getHeight(),true) and bullet.fromPlayer == false then
          -- if one of the enemy's bullets touches the player
          table.remove(bullets,j)
          if not(curPowerup == "shield" and isPowerup) then
            player.health = player.health-1
            love.audio.play(randElement(hurtSounds))
          end
        end
        if bullet.x > love.graphics.getWidth() or bullet.x < 0 or bullet.y > love.graphics.getHeight() or bullet.y < 0 then
          --if the bullet goes off screen then it will be removed
          table.remove(bullets,j)
        end
      end
      for i,box in pairs(healthBoxes) do
        if checkBBCollision(player.x,player.y,player.img:getWidth(),player.img:getHeight(),box.x,box.y,hlthBoxImg:getWidth(),hlthBoxImg:getHeight()) then
          -- if player touches a health box
          player.health = player.health + math.random(1,5)
          table.remove(healthBoxes,i)
          love.audio.play(healSound)
        end
      end
      player.angle = math.atan2(love.mouse.getY()-player.y,love.mouse.getX()-player.x)--makes player rotate towards the mouse
      if player.health > player.shopVals.maxHealth.value then player.health = player.shopVals.maxHealth.value end--keeps player's health being below maximum health
      if player.health <= 0 then
        --player dies here
        gameOver = true
        player.img = love.graphics.newImage("images/playership_broken.png")
      end
      if randomChance(1) and player.health <= player.shopVals.maxHealth.value - 5 and tableLength(healthBoxes) < 1 then
        --adds health box here
        table.insert(healthBoxes,{x=math.random(0,love.graphics.getWidth()),y=math.random(0,love.graphics.getHeight())})
      end
      for i,star in pairs(stars) do
        --controls "stars" here
        star.y = star.y + 0.5 * dt * timeMult * gameSpeed * worldSpeed
        if star.y > love.graphics.getHeight()+5 then
          star.y = -5
          star.x = math.random(0,love.graphics.getWidth())
        end
      end
      if powerupTimer < powerupTimerMax and not powerupAv then powerupTimer = powerupTimer + 0.02 * timeMult * dt * gameSpeed else
        powerupAv = true
      end
      if isPowerup and powerupTimer >= 0 then
        powerupTimer = powerupTimer - 0.05 * timeMult * dt * gameSpeed
        if powerupTimer <= 0 then
          isPowerup = false
          powerupTimer = 0
          worldSpeed = 1
          curPowerup = randElement(powerups)
        end
      end
      if powerupAv then
        if love.keyboard.isDown('space') then
          powerupAv = false
          isPowerup = true
        end
      end
      if isPowerup then
        if curPowerup == "slow" then worldSpeed = 0.2 end
      end
    end
    if love.keyboard.isDown('p') then powerupTimer = powerupTimerMax - 1 end
    if love.keyboard.isDown('o') then powerupTimer = 1 end
  end
  if love.keyboard.isDown('lshift','rshift') and love.keyboard.isDown('escape') then love.event.quit() end
end
function love.keypressed(key)
  if gameStarted then
    if key == 'r' then
      --restarts game here
      gameOver = false
      enemies = {}
      bullets = {}
      player.bullets = {}
      healthBoxes = {}
      score = 0
      player.health = player.shopVals.maxHealth.value
      player.x = love.graphics.getWidth()/2
      player.y = love.graphics.getHeight()/2
      player.img = love.graphics.newImage("images/playership.png")
      addEnemyTimer = 0
      powerupTimer = 0
      powerupAv = false
      isPowerup = false
      curPowerup = randElement(powerups)
      gameSpeed = 1
      worldSpeed = 1
      stars = {}
      for i=1,50 do
        table.insert(stars,{x=math.random(0,love.graphics.getWidth()),y=math.random(0,love.graphics.getHeight())})--make "stars"
      end
    end
    if key == 'escape' then gamePaused = not gamePaused end
  else
    if key =='return' then
      gameStarted = true
      love.mouse.setPosition(player.x,player.y)
    end
  end
end
function love.mousepressed(x,y,button)
  if not gameOver then
    --makes player shoot (when the mouse clicks) here
    if button == 1 then player.shootTimer = player.shootTimerMax end
  end
end
function love.draw(dt)
  if gameStarted then
    local shakeX = math.random(-2,2)
    local shakeY = math.random(-2,2)
    love.graphics.setColor(255,255,255,255)
    if player.health < player.shopVals.maxHealth.value/3 then love.graphics.translate(shakeX,shakeY) end
    for i,star in pairs(stars) do
      love.graphics.setColor(150,150,150)
      love.graphics.circle("fill",star.x,star.y,5)
    end
    love.graphics.setColor(255,255,255,255)
    if curPowerup == "shield" and isPowerup then love.graphics.draw(shieldImg,player.x,player.y,0,1,1,shieldImg:getWidth()/2,shieldImg:getHeight()/2) end
    love.graphics.draw(player.img,player.x,player.y,player.angle,1,1,player.img:getWidth()/2,player.img:getHeight()/2)
    for i,enemy in pairs(enemies) do
      love.graphics.draw(enemyImg,enemy.x,enemy.y,enemy.angle,1,1,enemyImg:getWidth()/2,enemyImg:getHeight()/2)
      love.graphics.setColor(255,0,0,150)
      love.graphics.rectangle("fill",enemy.x-enemyImg:getWidth()/2,enemy.y-enemyImg:getHeight()/2,enemyImg:getWidth(),13)
      love.graphics.setColor(0,255,0,150)
      love.graphics.rectangle("fill",enemy.x-enemyImg:getWidth()/2,enemy.y-enemyImg:getHeight()/2,(enemyImg:getWidth()/enemy.maxHealth)*enemy.health,13)
      love.graphics.setColor(255, 255, 255, 255)
      love.graphics.print(enemy.health .. "/" .. enemy.maxHealth,enemy.x,enemy.y,0,1,1,enemyImg:getWidth()/2,enemyImg:getHeight()/2)
    end
    for i,bullet in pairs(bullets) do
      love.graphics.setColor(bullet.c.r,bullet.c.g,bullet.c.b)
      love.graphics.draw(bulletImg,bullet.x,bullet.y,bullet.angle,1,1,bulletImg:getWidth()/2,bulletImg:getHeight()/2)
    end
    for i,box in pairs(healthBoxes) do
      love.graphics.setColor(255,255,255)
      love.graphics.draw(hlthBoxImg,box.x,box.y,0,1,1,hlthBoxImg:getWidth()/2,hlthBoxImg:getHeight()/2)
    end
    love.graphics.setColor(255,0,0,150)
    love.graphics.rectangle("fill",player.x-player.img:getWidth()/2,player.y-player.img:getHeight()/2,player.img:getWidth(),13)
    love.graphics.setColor(0,255,0,150)
    love.graphics.rectangle("fill",player.x-player.img:getWidth()/2,player.y-player.img:getHeight()/2,(player.img:getWidth()/player.shopVals.maxHealth.value)*player.health,13)
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.print(player.health .. "/" .. player.shopVals.maxHealth.value,player.x,player.y,0,1,1,player.img:getWidth()/2,player.img:getHeight()/2)
    love.graphics.setColor(200,200,255,100)
    love.graphics.setLineWidth(3)
    love.graphics.line(player.x,player.y,player.x+math.cos(player.angle)*(love.graphics.getWidth()*2),player.y + math.sin(player.angle)*(love.graphics.getHeight()*2))
    love.graphics.setColor(255, 255, 255, 255)
    if gameOver then
      love.graphics.print("GAME OVER!\nPRESS R TO RESTART",love.graphics.getWidth()/2,love.graphics.getHeight()/2)
    end
    love.graphics.setColor(255,255,255,255)
    if not gameOver then
      if not gamePaused then love.graphics.draw(cursorImg,love.mouse.getX(),love.mouse.getY(),player.angle,1,1,cursorImg:getWidth(),cursorImg:getHeight()/2) else love.graphics.draw(cursorImg,love.mouse.getX(),love.mouse.getY(),-1.5707963268,1,1,cursorImg:getWidth(),cursorImg:getHeight()/2) end
    end
    if player.health < player.shopVals.maxHealth.value/3 then love.graphics.translate(-shakeX,-shakeY) end
    love.graphics.print("Score: " .. score)
    if gamePaused then
      love.graphics.setColor(0,0,0,50)
      love.graphics.rectangle("fill",0,0,love.graphics.getWidth(),love.graphics.getHeight())
      love.graphics.setColor(255,255,255,255)
      love.graphics.print("Game Paused.\nPress ESCAPE to resume.\nPress SHIFT+ESCAPE to quit.",love.graphics.getWidth()/2,love.graphics.getHeight()/2)
    end
    love.graphics.rectangle("fill",0,400,20,(powerupTimer*(-200/powerupTimerMax)))
    if powerupAv then love.graphics.setColor(0,255,255) end
    love.graphics.rectangle("line",0,200,20,200)
    if not powerupAv and not isPowerup then love.graphics.print("Next powerup: " .. powerupsStrings[curPowerup],0,410) elseif powerupAv then love.graphics.print("\nPress SPACE\nto activate " .. powerupsStrings[curPowerup] .. "!",0,410) elseif not powerupAv and isPowerup then love.graphics.print(powerupsStrings[curPowerup] .. " is activated!",0,410) end
    love.graphics.setColor(255,255,255,255)
  else
    love.graphics.print("My Shooter Game! YAAY!!\n\nPress ENTER to start the game! \n\nAfter starting, use WASD or ARROW KEYS to move, move your MOUSE to where you want to\nshoot, and click the LEFT MOUSE BUTTON to shoot.")
  end
end